/*
 *				**** PSPAGES ****
 *
 * (c) 1992 Bill Pringle - wrp@ivy.unisys.com
 *
 * You may not sell this program, but you can charge a small
 * copying fee for distribution.  Permission granted for use
 * of this program for non-commercial use.  Commercial users
 * should contact me at the above e-mail address before use.
 *
 * This program prints certain pages from a postscript file.
 *
 * It assumes that standard page header comments are in the file.
 *
 */

#include <stdio.h>
#include <string.h>

#ifdef __MSDOS__
#include <stdlib.h>
#define	OPEN_RO	"rb"
#define	OPEN_RW	"wb"
#define	DIR_SLASH	'\\'
#else
#include <ctype.h>
#define SEEK_SET 0
#define	OPEN_RO	"r"
#define	OPEN_RW	"w"
#define	DIR_SLASH	'/'
#endif

#define PAGES	20		/* max. number of page groups */

#define	PRT_EVEN	1	/* print even pages */
#define	PRT_ODD		2	/* print odd pages */
#define	PRT_LOG		4	/* print logical pages */
#define	PRT_PHY		8	/* print physical pages */

struct Page
{		/* keep in-core list of page locations */
	struct	Page	*pg_next;		/* next entry */
	long	pg_beg;					/* start location of page */
	long	pg_end;					/* end location of page */
};

int		printdef=0;		/* default is odd pages */
int		revflg = 0;		/* default is in ascending order */
char	*Prog;			/* program name */
char	buf[BUFSIZ];	/* input buffer */
struct	Page	*Pages;	/* list of pages to be printed later */
long	Trailer;		/* loc of the trailer (if any needed) */

main(argc,argv)
int		argc;
char	**argv;
{
	FILE	*ifp;		/* input file */
	FILE	*ofp;		/* output file */
	int		printsw;	/* print switch */
	char	*p;			/* scan pointer */

	int	nent;				/* number of entries used */
	int		pglo[PAGES];	/* low page numbers */
	int		pghi[PAGES];	/* high page numbers */

	/* get  program name */

	if ((Prog = strrchr(argv[0],DIR_SLASH)) == NULL) Prog = argv[0];
	else Prog++;		/* skip directory slash */

#ifdef __MSDOS__
	for (p=Prog; *p != '\0'&& *p != '\.'; p++)
		if (isupper(*p)) *p = tolower(*p);
	*p = '\0';
#endif

	argc--;
	argv++;

	ifp = stdin;		/* default input file */
	ofp = stdout;		/* default output file */
	Pages = NULL;		/* no pages yet */
	Trailer = 0L;		/* no trailer either */

	if (argc == 0) { do_help(Prog); exit(0); }

	for (; argc > 0; argc--, argv++)
	{		/* process arguments */
		if (argv[0][0] == '-' && argv[0][1] != '\0') switch(argv[0][1])
		{		/* process option */
			case 'e':	/* even pages */
				printdef = PRT_EVEN;		/* print even pages */
				break;

			case 'f':	/* output file */
				if (argv[0][2] == '\0')
				{		/* file name is next argument */
					argc--;
					argv++;
					p = argv[0];			/* file name */
				} else
				{		/* file name follows -f */
					p = argv[0] + 2;
				}

				if ((ofp = fopen(p,OPEN_RW)) == NULL)
				{		/* problem opening file */
					fprintf(stderr,"%s: Can't open output file %s.\n",
								Prog, p);
					exit(3);
				}
				break;

			case 'h':	/* help */
				do_help(Prog);
				exit(0);

			case 'l':	/* logical pages */
				printdef = PRT_LOG;		/* print logical pages */
				if (argv[0][2] == '\0')
				{		/* pages are next argument */
					argc--;
					argv++;
					p = argv[0];
				} else p = (argv[0]) + 2;

				nent = parse_nums(p,pglo,pghi,PAGES);		/* get the actual page numbers */
				break;

			case 'o':	/* odd pages */
				printdef = PRT_ODD;		/* print odd pages */
				break;

			case 'p':	/* physical pages */
				printdef = PRT_PHY;		/* print physical pages */
				if (argv[0][2] == '\0')
				{		/* pages are next argument */
					argc--;
					argv++;
					p = argv[0];
				} else p = (argv[0]) + 2;

				nent = parse_nums(p,pglo,pghi,PAGES);
				break;

			case 'r':	/* print in reverse order */
				revflg++;		/* signal reverse */
				break;

			default:	/* bad argument */
				fprintf(stderr,"%s: Invalid argument (%s)\n", Prog, argv[0]);
				do_usage(Prog);
				exit(1);
		} else
		{		/* file name */
			if (argv[0][0] == '-') ifp = stdin;
			else if ((ifp = fopen(argv[0],OPEN_RO)) == NULL)
			{		/* bad file */
				fprintf(stderr,"%s: Can't open file %s for input.\n",
									Prog, argv[0]);
				exit(2);
			}

			copy_prologue(ifp,ofp);

			if (printdef == PRT_EVEN || printdef == PRT_ODD)
			{		/* either even or odd given */
				if (printdef == PRT_EVEN) printsw = ~0;
				else printsw = 0;

				even_odd(ifp,ofp,printsw);
			} else pspages(ifp,ofp,printdef,nent,pglo,pghi);

			if (revflg) revpgs(ifp,ofp);		/* print pages in reverse */

			if (ifp != stdin) fclose(ifp);		/* close input */
		}
	}
	if (ofp != stdout) fclose(ofp);		/* close output */

	fprintf(stderr,"%s: Program terminating.\n", Prog);
	exit(0);
}

copy_prologue(ifp,ofp)
FILE	*ifp;
FILE	*ofp;
{		/* copy prologue */
	long	loc=0L;		/* location in file */
	long	page1;		/* location of first page */

	while (fgets(buf,sizeof(buf),ifp) != NULL)
	{		/* read entire file */
		if (stribeg(buf,"%%page: ") == 0) break;
		fputs(buf,ofp);		/* write out line */
		loc = ftell(ifp);	/* location of next line */
	}

	if (stribeg(buf,"%%page: "))
	{		/* no pages found */
		fprintf(stderr,"%s: Input file not Postscript Conforming\n", Prog);
		exit(5);
	} else page1 = loc;		/* save location of first page */

	while (fgets(buf,sizeof(buf),ifp) != NULL)
	{		/* look for trailer */
		if (stribeg(buf,"%%trailer") == 0) Trailer = loc;
		if (stribeg(buf,"%%page: ") == 0) Trailer = 0L;	/* page trailer */
		loc = ftell(ifp);	/* loc of next line */
	}

	if (fseek(ifp,page1,SEEK_SET))		/* reposition to just before %%Page: */
	{		/* seek failed */
		printf("%s: Unable to seek loc %ld after prolog copied\n", Prog, loc);
		exit(4);
	}
}

even_odd(ifp,ofp,sw)
FILE	*ifp;			/* input file */
FILE	*ofp;			/* output file */
int		sw;				/* print page switch */
{		/* print even or odd pages */
	long	loc;			/* location in file */

	loc = ftell(ifp);		/* location in file */

	while (fgets(buf,sizeof(buf),ifp) != NULL)
	{		/* read entire file */
		if (stribeg(buf,"%%page: ") == 0)
		{		/* new page - toggle switch & save loc if needed */
			sw = ~ sw;	/* toggle switch */
			if (revflg)
			{		/* printing in reverse order - save locations */
				if (sw) begpage(loc);	/* loc of this line */
				else endpage(ftell(ifp)); /* loc of next line */
			}
		}

		loc = ftell(ifp);		/* location in file */

		if (sw && !revflg)
		{		/* printing this page in forward direction */
			fputs(buf,ofp);		/* print line if needed */
		}
	}

	/* done with file - save last location if needed */
	if (revflg && sw)
	{		/* printing later - get end of file */
		endpage(ftell(ifp)+1L);	/* when to terminate */
	}
}

pspages(ifp,ofp,prsw,nent,pglo,pghi)
FILE	*ifp;		/* input file */
FILE	*ofp;		/* output file */
int		prsw;		/* physical/logical page flag */
int		nent;		/* number of page groups */
int		pglo[];		/* low page numbers */
int		pghi[];		/* high page numbers */
{
	int		sw=0;			/* print switch */
	long	loc;			/* location in file */

	loc = ftell(ifp);		/* location of first line */

	while (fgets(buf,sizeof(buf),ifp) != NULL)
	{		/* read entire file */

		if (stribeg(buf,"%%page: ") == 0)
		{		/* new page found - check it out */
			if (sw && revflg) endpage(loc);
			sw = chk_pg(buf,prsw,nent,pglo,pghi);
			if (sw && revflg) begpage(loc);	/* save new page */
		}
		loc = ftell(ifp);		/* get location of next page */

		if (sw & !revflg) fputs(buf,ofp);		/* print line if needed */
	}

	/* done with file - save last location if needed */
	if (revflg && sw) endpage(ftell(ifp));
}

chk_pg(str,prsw,nent,pglo,pghi)
char	*str;		/* line with page number */
int		prsw;		/* physical/logical page flag */
int		nent;		/* number of page groups */
int		pglo[];		/* low page numbers */
int		pghi[];		/* high page numbers */
{
	int		pgno;	/* page number */
	int		i;		/* loop variable */

	while (*str != '\0' && ! isspace(*str)) str++;	/* find space */
	while (isspace(*str)) str++;	/* find logical page number */

	if (prsw == PRT_PHY)
	{		/* printing physical pages - get number */
		while (*str != '\0' && !isspace(*str)) str++;	/* end of logical */
		while (isspace(*str)) str++;					/* start of physical */
	}

	pgno = atoi(str);		/* get page number */

	for (i = 0; i < nent; i++)
	{		/* loop thru selected page numbers */
		if (pgno >= pglo[i] && pgno <= pghi[i]) return 1;
	}

	return 0;
}

get_line(buf,fp)
char	*buf;
FILE	*fp;
{
	int		c, n;

	for (n = 0; (c = fgetc(fp)) != EOF; n++)
	{		/* read up to end of line */
		if (c !=  '\r' && c != '\n') *buf++ = c;
		else if (c == '\n') break;
	}

	*buf = '\0';		/* termiate line */
	return n;
}

stribeg(str1,str2)
char	*str1, *str2;
{
	size_t	len;

	if ((len = strlen(str2)) == 0) return 0;
	else return (strnicmp(str1,str2,len));
}

do_help(Prog)
char	*Prog;
{
	static char *msg[] = {
"This program will print portions of a postscript file.\n",
"Options are:\n",
"\t-e\tPrint only even physical pages.\n",
"\t-o\tPrint only odd physical pages.\n",
"\t-l pgs\tPrint logical pages 'pgs'\n",
"\t-p pgs\tPrint physical pages 'pgs'\n",
"\t-r\tPrint pages in reverse order\n",
"\n",
"Where 'pgs' is of the form: range1[,range2 ...]\n",
"\tFor example: 1-5,26,35-76\n",
"\n",
"To print d1208ouble sided, type:\n",
"\t%s -o infile.ps > front.ps\n",
"\t%s -e -r infile.ps > back.ps\n",
"Then print back.ps on the reverse side (holes on right)\n",
"turn paper over, add a blank sheet at the bottom (in case\n",
"an odd number of pages were printed) and print front.ps\n",
NULL};

	prt_msg(Prog,msg);
}

do_usage(Prog)
char	*Prog;
{
	static char *msg[] = {
"Usage: %s [-e | -o | -l pages | -p pages ] [ -r ] ps_files ...\n",
"-OR- for help, type:\n",
"\t%s -h\n",
NULL};

	prt_msg(Prog,msg);
}

prt_msg(pgm,msg)
char	*pgm;		/* program name */
char	*msg[];		/* message lines */
{
	int		i;

	for (i=0; msg[i] != NULL; i++) fprintf(stderr,msg[i],pgm);
}

get_nums() {   return 0; }

begpage(loc)
long	loc;		/* location within file */
{		/* record beginning of new page to be printed later */
	struct	Page	*pg;	/* page entry pointer */

	pg = (struct Page *) malloc(sizeof(struct Page));	/* get memory */
	pg->pg_next = Pages;		/* rest of list to follow */
	pg->pg_beg = loc;						/* beg loc of page */
	pg->pg_end = 0L;						/* no end yet */
	Pages = pg;				/* put at front of list */
}

endpage(loc)
long	loc;		/* location within file */
{		/* record end of page to be printed later */
	if (Pages != NULL) Pages->pg_end = loc;
}

revpgs(ifp,ofp)
FILE	*ifp;		/* input file */
FILE	*ofp;		/* output file */
{		/* print pages in reverse */
	struct	Page	*pg;		/* temp pointer */
	long	loc;				/* location in file */

	while (Pages !=  NULL)
	{		/* page still remainins */
		fseek(ifp,Pages->pg_beg,SEEK_SET);	/* start of page */
		while (fgets(buf,sizeof(buf),ifp) != NULL)
		{		/* still more to print */
			if ((loc = ftell(ifp)) >= Pages->pg_end) break;
			if (Trailer > 0L && loc > Trailer) break;	/* trailer */
			fputs(buf,ofp);					/* and write it */
		}

		pg = Pages->pg_next;				/* new top of stack */
		free(Pages);						/* free top element */
		Pages = pg;						/* make new top */
	}

	if (Trailer != 0L)
	{		/* print trailer */
		fseek(ifp,Trailer,SEEK_SET);	/* start of trailer */
		while (fgets(buf,sizeof(buf),ifp) != NULL) fputs(buf,ofp);
	}
}
